/* PhyloPattern
 * Copyright (C) 2005-2009 Philippe Gouret
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA. The text ol license can be also found 
 * at http://www.gnu.org/copyleft/lgpl.html
 */

//------------------------------------------------------------------------------
// Caller environment Implementation
// Philippe Gouret - January 2002
//------------------------------------------------------------------------------

/**
 *  Signal
 *  @since JDK1.2
 */
package gnu.prolog.test;

//imported library classes
import java.util.*;
import java.io.*;

//import project classes
import gnu.prolog.vm.*;
import gnu.prolog.term.*;
import gnu.prolog.io.*;
import gnu.prolog.database.PrologTextLoaderError;

/**
 *  Caller environment Implementation class
 *  @author Philippe Gouret
 */
public class CallerContext
{
	//----------------------------------------------------------------------------
	// Constants
	//----------------------------------------------------------------------------

	/** predicate call is successfull with multiple solution */
	static final public short CALL_SUCCESSMULTIPLE = 1;
	/** predicate call is successfull */
	static final public short CALL_SUCCESSFULL = 0;
	/** predicate call failed */
	static final public short CALL_FAILED = -1;
	/** error on predicate call */
	static final public short CALL_ERROR = -2;

	//----------------------------------------------------------------------------
	// Public members
	//----------------------------------------------------------------------------

	//----------------------------------------------------------------------------
	// Private members
	//----------------------------------------------------------------------------

	// caller uid
	private String uid;

	// Prolog environment
	private gnu.prolog.vm.Environment prologEnvironment;

	// Prolog interpreter
	private Interpreter prologInterpreter;

	// current goal
	private HashMap goalList = new HashMap();

	// current goal read options
	private HashMap readOptionsList = new HashMap();

	//----------------------------------------------------------------------------
	// Private methods
	//----------------------------------------------------------------------------

	/**
	 *  load Prolog file
	 *  @param filename
	 */
	private void loadPrologFile(String filename)
	{
		// try to load file
		prologEnvironment.ensureLoaded(AtomTerm.get(filename));

		// display errors if any
		for (Iterator errors = prologEnvironment.getLoadingErrors().iterator(); errors.hasNext();)
			System.err.println((PrologTextLoaderError) errors.next());
	}

	//----------------------------------------------------------------------------
	// Public methods
	//----------------------------------------------------------------------------

	/**
	 *  get uid
	 *  @return it
	 */
	public String getUID()
	{
		return uid;
	}

	/**
	 *  test equality
	 *  @param context
	 */
	public boolean equals(Object o)
	{
		// compare uid
		return uid.equals(((CallerContext) o).getUID());
	}

	/**
	 *  stop current goal
	 *  @param request id
	 */
	synchronized public void stopCurrentGoal(String requestId)
	{
		// get current goal
		Interpreter.Goal goal = (Interpreter.Goal) goalList.get(requestId);
		ReadOptions readOptions = (ReadOptions) readOptionsList.get(requestId);
		if (goal != null)
		{
			prologInterpreter.stop(goal);
			goalList.remove(goal);
			readOptionsList.remove(readOptions);
		}
	}

	/**
	 *  get results names (variables) for last goal
	 *  @param request id
	 */
	synchronized public Vector getResultsNames(String requestId)
	{
		// get read options
		ReadOptions readOptions = (ReadOptions) readOptionsList.get(requestId);
		if (readOptions == null) return null;

		// read variables names
		Vector results = new Vector();
		Iterator variables = readOptions.variableNames.keySet().iterator();
		while (variables.hasNext()) results.add(variables.next());
		return results;
	}

	/**
	 *  get results for last goal
	 *  @param request id
	 */
	synchronized public Vector getResults(String requestId)
	{
		// get read options
		ReadOptions readOptions = (ReadOptions) readOptionsList.get(requestId);
		if (readOptions == null) return null;

		// read variables
		Vector results = new Vector();
		Iterator variables = readOptions.variableNames.keySet().iterator();
		while (variables.hasNext())
		{
			// read variables as string
			Term varTerm = ((Term) readOptions.variableNames.get(variables.next())).dereference();
			String varStr;
			try
			{
				varStr = TermWriter.toString(varTerm);
			}
			catch (Exception e)
			{
				varStr = "";
			}

			results.add(varStr);
		}
		return results;
	}

	/**
	 *  call a predicate
	 *  @param request id
	 *  @param predicate to call
	 */
	synchronized public short callPredicate(String requestId, String predicate)
	{
		try
		{
			// new goal and read options
			Interpreter.Goal goal = null;
			ReadOptions readOptions;

			// new predicate
			if (!predicate.equals(""))
			{
				// build predicate term
				TermReader termReader = new TermReader(new StringReader(predicate));
				readOptions = new ReadOptions();
				readOptions.operatorSet = prologEnvironment.getOperatorSet();
				Term goalTerm = termReader.readTermEof(readOptions);

				// ask interpreter to run the goal
				goal = prologInterpreter.prepareGoal(goalTerm);

				// keep current goal
				goalList.put(requestId, goal);
				readOptionsList.put(requestId, readOptions);
			}
			else
			{
				// get in current lists
				goal = (Interpreter.Goal) goalList.get(requestId);
				readOptions = (ReadOptions) readOptionsList.get(requestId);
			}

			// no current goal
			if (goal == null) return CALL_ERROR;

			// execute
			switch(prologInterpreter.execute(goal))
			{
			// successfull with many solutions
			case PrologCode.SUCCESS:
				return CALL_SUCCESSMULTIPLE;

				// successfull with one solution
			case PrologCode.SUCCESS_LAST:
				return CALL_SUCCESSFULL;

				// failed
			case PrologCode.FAIL:
				return CALL_FAILED;

				// unknown
			default:
				return CALL_ERROR;
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();

			// syntax error or unknown predicate or ...
			return CALL_ERROR;
		}
	}

	/**
	 *  dummy constructor used for searching
	 *  @param uid
	 */
	public CallerContext(String uid)
	{
		// keep info
		this.uid = uid;
	}

	/**
	 *  constructor
	 *  @param uid
	 *  @param file list
	 */
	public CallerContext(String uid, List<String> fileList)
	{
		// keep info
		this.uid = uid;

		// initialize Prolog environment and interpreter
		prologEnvironment = new gnu.prolog.vm.Environment();
		prologInterpreter = prologEnvironment.createInterpreter();

		// search files to load
		for (String filename: fileList) loadPrologFile(filename);
	}
}


